/*-
 * Copyright (c) 2015  Peter Pentchev
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <assert.h>
#include <stdbool.h>
#include <dirent.h>
#include <err.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef __printflike
#if defined(__GNUC__) && __GNUC__ >= 3
#define __printflike(x, y)	__attribute__((format(printf, (x), (y))))
#else
#define __printflike(x, y)
#endif
#endif

static int		verbose;

static void		usage(int _ferr);
static void		version(void);
static void		debug(const char *fmt, ...) __printflike(1, 2);

static int		check_wait_result(pid_t res, int stat, pid_t expected, const char *progname);
static char		*get_pattern(char *pattern);
static const char	*find_temploc(void);
static void		rmdir_recursive(const char *dirname);

static void		signal_handler(int);

static volatile pid_t	childpid = -1;

int
main(int argc, char * const argv[])
{
	int ch, hflag, Vflag, noop;
	const char *varname;
	char *pattern_arg;

	hflag = noop = Vflag = 0;
	varname = pattern_arg = NULL;
	while (ch = getopt(argc, argv, "e:hNt:Vv"), ch != -1)
		switch (ch) {
			case 'e':
				varname = optarg;
				break;

			case 'h':
				hflag = 1;
				break;

			case 'N':
				noop = 1;
				break;

			case 'V':
				Vflag = 1;
				break;

			case 't':
				pattern_arg = optarg;
				break;

			case 'v':
				verbose++;
				break;

			default:
				usage(1);
				/* NOTREACHED */
		}
	if (Vflag)
		version();
	if (hflag)
		usage(0);
	if (Vflag || hflag)
		return (0);

	argc -= optind;
	argv += optind;
	if (argc < 1) {
		warnx("No command to run");
		usage(1);
	} else if (varname == NULL) {
		warnx("No variable name specified using the -e option");
		usage(1);
	}

	debug("Setting up the signal handlers\n");
	struct sigaction act;
	memset(&act, 0, sizeof(act));
	act.sa_handler = signal_handler;
	act.sa_flags = SA_RESTART;

	if (sigaction(SIGHUP, &act, NULL) == -1)
		err(1, "Could not set the SIGHUP handler");
	if (sigaction(SIGINT, &act, NULL) == -1)
		err(1, "Could not set the SIGINT handler");
	if (sigaction(SIGQUIT, &act, NULL) == -1)
		err(1, "Could not set the SIGQUIT handler");
	if (sigaction(SIGUSR1, &act, NULL) == -1)
		err(1, "Could not set the SIGUSR1 handler");
	if (sigaction(SIGUSR2, &act, NULL) == -1)
		err(1, "Could not set the SIGUSR2 handler");
	if (sigaction(SIGALRM, &act, NULL) == -1)
		err(1, "Could not set the SIGALRM handler");
	if (sigaction(SIGTERM, &act, NULL) == -1)
		err(1, "Could not set the SIGTERM handler");
	if (sigaction(SIGCONT, &act, NULL) == -1)
		err(1, "Could not set the SIGCONT handler");

	debug("Creating the temporary directory\n");
	char * const pattern = get_pattern(pattern_arg);
	assert(pattern != NULL);
	debug("- pattern: %s\n", pattern);
	char * const tempdir = mkdtemp(pattern);
	if (tempdir == NULL)
		err(1, "Could not create a directory as %s", pattern);
	debug("- directory: %s\n", tempdir);
	if (setenv(varname, tempdir, 1) == -1) {
		warn("Could not set the %s variable to %s", varname, tempdir);
		if (rmdir(tempdir) == -1)
			err(1, "Could not remove the (empty, just created) "
			    "temporary directory %s", tempdir);
		exit(1);
	}

	int res;
	if (noop) {
		res = 0;
		goto REMOVE;
	}

	debug("Running the command (%s)\n", argv[0]);
	childpid = fork();
	const pid_t pid = childpid;
	if (pid == -1)
		err(1, "Could not fork for %s", argv[0]);
	else if (pid == 0) {
		execvp(argv[0], argv);
		err(1, "Could not execute %s", argv[0]);
	}
	debug("- child pid %jd\n", (intmax_t)pid);

	debug("Waiting for the command to complete\n");
	int status;
	const pid_t npid = waitpid(pid, &status, 0);
	childpid = -1;
	debug("- exit status for pid %jd: %d\n", (intmax_t)npid, status);
	res = check_wait_result(npid, status, pid, argv[0]);
	debug("- figured out our exit status %d\n", res);

REMOVE:
	debug("Removing the temporary directory\n");
	rmdir_recursive(tempdir);
	return (res);
}

void
usage(int _ferr)
{
	const char *s =
	    "Usage:\ttempdir [-Nv] [-t template] -e varname command args...\n"
	    "\ttempdir -V | -h\n"
	    "\n"
	    "\t-e\tspecify the name of the variable to store the path in\n"
	    "\t-h\tdisplay program usage information and exit\n"
	    "\t-N\tno operation mode; display what would have been done\n"
	    "\t-t\tspecify a pattern for the temporary directory name\n"
	    "\t-V\tdisplay program version information and exit\n"
	    "\t-v\tverbose operation; display diagnostic output\n";

	if (_ferr)
		errx(1, "%s", s);
	else
		printf("%s", s);
}

void
version(void)
{
	puts("tempdir 0.1.0");
}

void
debug(const char *fmt, ...)
{
	va_list v;

	va_start(v, fmt);
	if (verbose)
		vfprintf(stderr, fmt, v);
	va_end(v);
}

int
check_wait_result(const pid_t pid, const int stat, const pid_t expected,
	const char * const progname)
{
	if (pid != expected) {
		warnx("Waiting for %s: expected pid %jd, got %jd", progname,
		    (intmax_t)expected, (intmax_t)pid);
		return (1);
	} else if (WIFEXITED(stat)) {
		const int code = WEXITSTATUS(stat);

		if (code != 0)
			warnx("Child %s (pid %jd) exited with code %d",
			    progname, (intmax_t)pid, code);
		return (code);
	} else if (WIFSIGNALED(stat)) {
		const int sig = WTERMSIG(stat);

		warnx("Child %s (pid %jd) was killed by signal %d", progname,
		    (intmax_t)pid, sig);
		return (sig > 127? 255: sig + 128);
	} else if (WIFSTOPPED(stat)) {
		const int sig = WSTOPSIG(stat);

		warnx("Child %s (pid %jd) was stopped by signal %d", progname,
		    (intmax_t)pid, sig);
		return (sig > 127? 255: sig + 128);
	} else {
		warnx("Child %s (pid %jd) neither exited nor was killed nor "
		    "stopped; what in the world does wait(2) status %d mean?!",
		    progname, (intmax_t)pid, stat);
		return (1);
	}
	/* NOTREACHED */
}

#define TEMPSUFFIX	"/tempdir.XXXXXX"

char *
get_pattern(char * const pattern_arg)
{
	if (pattern_arg != NULL) {
		const size_t len = strlen(pattern_arg);

		if (len < 6 || strcmp(pattern_arg + len - 6, "XXXXXX") != 0)
			errx(1, "The pattern must end in six 'X' characters");
		return (pattern_arg);
	}

	const char * const temploc = find_temploc();
	assert(temploc != NULL);
	const size_t temploclen = strlen(temploc);
	// sizeof() will leave room for the terminating NUL character
	const size_t patlen = temploclen + sizeof(TEMPSUFFIX);
	if (patlen >= INT_MAX)
		errx(1, "The path to the temporary directory is longer than "
		    "%d characters, this is completely unreasonable and "
		    "there is no way to detect errors", INT_MAX);
	char * const pattern = malloc(patlen);
	if (pattern == NULL)
		errx(1, "Out of memory");
	if (snprintf(pattern, patlen, "%s%s", temploc, TEMPSUFFIX) !=
	    (int)patlen - 1)
		errx(1, "snprintf() failed in some weird and wonderful way");
	return (pattern);
}

const char *
find_temploc(void)
{
	const char * const paths[] = {
		"$TMPDIR",
		"$TMP",
		"$TEMPDIR",
		"$TEMP",
		"/tmp",
		"/var/tmp",
		".",
	};
	for (unsigned i = 0; i < sizeof(paths) / sizeof(paths[0]); i++) {
		const char * const path = paths[i][0] == '$'?
		    getenv(paths[i] + 1): paths[i];
		struct stat sbuf;
		
		if (path != NULL && stat(path, &sbuf) == 0 &&
		    S_ISDIR(sbuf.st_mode) && access(path, W_OK) == 0)
			return (path);
	}
	errx(1, "Could not find a suitable temporary directory location");
}

static void
rmdir_recursive(const char * const dirname)
{
	DIR * const d = opendir(dirname);
	
	if (d == NULL) {
		if (errno == ENOENT)
			return;
		err(1, "Could not open the %s directory", dirname);
	}

	debug("About to recursively remove directory %s\n", dirname);
	if (chmod(dirname, 0700) == -1)
		err(1, "Could not change the access permissions on directory %s", dirname);

	bool found;
	do {
		found = false;
		rewinddir(d);

		for (struct dirent *e = readdir(d); e != NULL; e = readdir(d)) {
			if (strcmp(e->d_name, ".") == 0 ||
			    strcmp(e->d_name, "..") == 0)
				continue;

			const size_t len = strlen(dirname) + 1 +
			    strlen(e->d_name) + 1;
			if (len >= INT_MAX)
				errx(1, "Path too long to build and remove: "
				    "%s/%s", dirname, e->d_name);
			char * const fname = malloc(len);
			if (fname == NULL)
				errx(1, "Out of memory");
			if (snprintf(fname, len, "%s/%s", dirname, e->d_name)
			    != (int)len - 1)
				errx(1, "snprintf(%s/%s) failed in some "
				    "weird and wonderful way",
				    dirname, e->d_name);
			found = true;

			struct stat sbuf;
			if (lstat(fname, &sbuf) == -1)
				err(1, "Could not stat %s", fname);
			if (S_ISDIR(sbuf.st_mode)) {
				rmdir_recursive(fname);
			} else {
				debug("About to remove non-directory %s\n",
				    fname);
				if (!S_ISLNK(sbuf.st_mode) &&
				    chmod(fname, 0600) == -1)
					err(1, "Could not change the access "
					    "permissions on non-directory %s",
					    fname);
				if (unlink(fname) == -1)
					err(1, "Could not remove "
					    "non-directory %s", fname);
			}

			free(fname);
		}
	} while (found);

	if (rmdir(dirname) == -1)
		err(1, "Could not remove the supposedly empty %s directory",
		    dirname);
}

static void
signal_handler(int sig)
{
	const pid_t pid = childpid;
	if (pid == -1)
		return;

	if (verbose) {
		char buf[] = "Forwarding signal NN\n";
#define szbuf	sizeof(buf)

		size_t len = szbuf - 1;
		if (sig < 10) {
			buf[szbuf - 4] = sig + '0';
			buf[szbuf - 3] = '\n';
			len--;
		} else {
			buf[szbuf - 4] = (sig / 10) + '0';
			buf[szbuf - 3] = (sig % 10) + '0';
		}

		size_t written = 0;
		while (written < len) {
			ssize_t n = write(2, buf + written, len - written);
			if (n < 1)
				break;
			written += n;
		}
	}

	kill(pid, sig);
}
